@@ -2,13 +2,16 @@
 '''Command line arguments parser
 '''
 
-import sys, traceback, getopt, types, textwrap, inspect, os, copy
+import sys, traceback, getopt, types, textwrap, inspect, os, copy, keyword
 from itertools import imap
+from functools import wraps
 
-__all__ = ['command', 'dispatch']
-__version__ = '2.1'
+
+__all__ = ['Dispatcher', 'command', 'dispatch']
+__version__ = '3.0'
 __author__ = 'Alexander Solovyov'
-__email__ = 'piranha@piranha.org.ua'
+__email__ = 'alexander@solovyov.net'
+
 
 try:
     import locale
@@ -29,153 +32,205 @@ def err(text):
     '''Write output to stderr'''
     write(text, out=sys.stderr)
 
-CMDTABLE = {}
 
-# --------
-# Public interface
-# --------
+class Dispatcher(object):
+    '''Central object for command dispatching system
 
-def command(options=None, usage=None, name=None, shortlist=False, hide=False):
-    '''Decorator to mark function to be used for command line processing.
-
-    All arguments are optional:
-
-     - ``options``: options in format described later. If not supplied,
-       will be determined from function.
-     - ``usage``: usage string for function, replaces ``%name`` with name
-       of program or subcommand. In case if it's subcommand and ``%name``
-       is not present, usage is prepended by ``name``
-     - ``name``: used for multiple subcommands. Defaults to wrapped
-       function name
-     - ``shortlist``: if command should be included in shortlist. Used
-       only with multiple subcommands
-     - ``hide``: if command should be hidden from help listing. Used only
-       with multiple subcommands, overrides ``shortlist``
-
-    Options should be a list of 4-tuples in format::
-
-      (shortname, longname, default, help)
-
-    Where:
-
-     - ``shortname`` is a single letter which can be used then as an option
-       specifier on command line (like ``-a``). Will be not used if contains
-       falsy value (empty string, for example)
-     - ``longname`` - main identificator of an option, can be used as on a
-       command line with double dashes (like ``--longname``)
-     - ``default`` value for an option, type of it determines how option will be
-       processed
-     - ``help`` string displayed as a help for an option when asked to
-    '''
-    def wrapper(func):
-        try:
-            options_ = list(guess_options(func))
-        except TypeError:
-            options_ = []
-        try:
-            options_ = options_ + list(options)
-        except TypeError:
-            pass
-
-        name_ = name or func.__name__.replace('_', '-')
-        if usage is None:
-            usage_ = guess_usage(func, options_)
-        else:
-            usage_ = usage
-        prefix = hide and '~' or (shortlist and '^' or '')
-        CMDTABLE[prefix + name_] = (func, options_, usage_)
-
-        def help_func(name=None):
-            return help_cmd(func, replace_name(usage_, sysname()), options_)
-        func.help = help_func
-
-        @wraps(func)
-        def inner(*args, **opts):
-            # look if we need to add 'help' option
-            try:
-                (True for option in reversed(options_)
-                 if option[1] == 'help').next()
-            except StopIteration:
-                options_.append(('h', 'help', False, 'show help'))
-
-            argv = opts.pop('argv', sys.argv[1:])
-            if opts.pop('help', False):
-                return func.help()
-
-            if args or opts:
-                # no catcher here because this is call from Python
-                return call_cmd_regular(func, options_)(*args, **opts)
-
-            try:
-                opts, args = catcher(lambda: parse(argv, options_), func.help)
-            except Abort:
-                return -1
-
-            if opts.pop('help', False):
-                return func.help()
-
-            try:
-                return catcher(lambda: call_cmd(name_, func)(*args, **opts),
-                               func.help)
-            except Abort:
-                return -1
-
-        return inner
-    return wrapper
-
-
-def dispatch(args=None, cmdtable=None, globaloptions=None,
-             middleware=lambda x: x):
-    '''Dispatch command arguments based on subcommands.
-
-    - ``args``: list of arguments, default: ``sys.argv[1:]``
-    - ``cmdtable``: dict of commands in format described below.
-      If not supplied, will use functions decorated with ``@command``.
+    - ``cmdtable``: dict of commands. Will be populated with functions,
+      decorated with ``Dispatcher.command``.
     - ``globaloptions``: list of options which are applied to all
       commands, will contain ``--help`` option at least.
     - ``middleware``: global decorator for all commands.
-
-    cmdtable format description::
-
-      {'name': (function, options, usage)}
-
-    - ``name`` is the name used on command-line. Can contain aliases
-      (separate them with ``|``), pointer to a fact that this command
-      should be displayed in short help (start name with ``^``), or to
-      a fact that this command should be hidden (start name with ``~``)
-    - ``function`` is the actual callable
-    - ``options`` is options list in format described in docs
-    - ``usage`` is the short string of usage
     '''
-    args = args or sys.argv[1:]
-    cmdtable = cmdtable or CMDTABLE
 
-    globaloptions = globaloptions or []
-    globaloptions.append(('h', 'help', False, 'display help'))
+    def __init__(self, cmdtable=None, globaloptions=None, middleware=None):
+        self._cmdtable = cmdtable or {}
+        self._globaloptions = globaloptions or []
+        self.middleware = middleware
 
-    cmdtable['help'] = (help_(cmdtable, globaloptions), [], '[TOPIC]')
-    help_func = cmdtable['help'][0]
+    @property
+    def globaloptions(self):
+        opts = self._globaloptions[:]
+        if not next((True for o in opts if o[1] == 'help'), None):
+            opts.append(('h', 'help', False, 'display help'))
+        return opts
 
-    autocomplete(cmdtable, args, middleware)
+    @property
+    def cmdtable(self):
+        table = self._cmdtable.copy()
+        table['help'] = help_(table, self.globaloptions), [], '[TOPIC]'
+        return table
 
-    try:
-        name, func, args, kwargs = catcher(
-            lambda: _dispatch(args, cmdtable, globaloptions),
-            help_func)
-    except Abort:
-        return -1
+    def command(self, options=None, usage=None, name=None, shortlist=False,
+                hide=False):
+        '''Decorator to mark function to be used as command for CLI.
 
-    if name == '_completion':       # skip middleware
-        worker = lambda: call_cmd(name, func)(*args, **kwargs)
+        Usage::
+
+          from opster import command, dispatch
+
+          @command()
+          def run(argument,
+                  optionalargument=None,
+                  option=('o', 'default', 'help for option'),
+                  no_short_name=('', False, 'help for this option')):
+              print argument, optionalargument, option, no_short_name
+
+          if __name__ == '__main__':
+              run.command()
+
+          # or, if you want to have multiple subcommands:
+          if __name__ == '__main__':
+              dispatch()
+
+        Optional arguments:
+         - ``options``: options in format described later. If not supplied,
+           will be determined from function.
+         - ``usage``: usage string for function, replaces ``%name`` with name
+           of program or subcommand. In case if it's subcommand and ``%name``
+           is not present, usage is prepended by ``name``
+         - ``name``: used for multiple subcommands. Defaults to wrapped
+           function name
+         - ``shortlist``: if command should be included in shortlist. Used
+           only with multiple subcommands
+         - ``hide``: if command should be hidden from help listing. Used only
+           with multiple subcommands, overrides ``shortlist``
+
+        If defined, options should be a list of 4-tuples in format::
+
+          (shortname, longname, default, help)
+
+        Where:
+
+         - ``shortname`` is a single letter which can be used then as an option
+           specifier on command line (like ``-a``). Will be not used if contains
+           falsy value (empty string, for example)
+         - ``longname`` - main identificator of an option, can be used as on a
+           command line with double dashes (like ``--longname``)
+         - ``default`` value for an option, type of it determines how option
+           will be processed
+         - ``help`` string displayed as a help for an option when asked to
+        '''
+        def wrapper(func):
+            try:
+                options_ = list(options or guess_options(func))
+            except TypeError:
+                options_ = []
+
+            name_ = name or name_from_python(func.__name__)
+            if usage is None:
+                usage_ = guess_usage(func, options_)
+            else:
+                usage_ = usage
+            prefix = hide and '~' or (shortlist and '^' or '')
+            self._cmdtable[prefix + name_] = (func, options_, usage_)
+
+            def help_func(name=None):
+                return help_cmd(func, replace_name(usage_, sysname()), options_)
+
+            def command(argv=None):
+                for o in self.globaloptions:
+                    if not next((x for x in options_ if
+                                 x[1] == o[1] or (x[0] and x[0] == o[0])),
+                                None):
+                        options_.append(o)
+
+                argv = argv or sys.argv[1:]
+                try:
+                    opts, args = parse(argv, options_)
+                except Exception, e:
+                    if exchandle(e, func.help):
+                        return -1
+                    raise
+
+                if opts.pop('help', False):
+                    return func.help()
+
+                try:
+                    return call_cmd(name_, func)(*args, **opts)
+                except Exception, e:
+                    if exchandle(e, func.help):
+                        return -1
+                    raise
+
+            func.usage = usage_
+            func.help = help_func
+            func.command = command
+            func.opts = options_
+            func.orig = func
+
+            @wraps(func)
+            def inner(*args, **opts):
+                return call_cmd_regular(func, options_)(*args, **opts)
+
+            return inner
+
+        return wrapper
+
+    def _dispatch(self, args):
+        cmd, func, args, options = cmdparse(args, self.cmdtable,
+                                            self.globaloptions)
+
+        if options.pop('help', False):
+            return 'help', self.cmdtable['help'][0], [cmd], {}
+        elif not cmd:
+            return 'help', self.cmdtable['help'][0], ['shortlist'], {}
+
+        return cmd, func, args, options
+
+    def dispatch(self, args=None):
+        '''Dispatch command line arguments using subcommands
+
+        - ``args``: list of arguments, default: ``sys.argv[1:]``
+        '''
+        args = args or sys.argv[1:]
+
+        help_func = self.cmdtable['help'][0]
+        autocomplete(self.cmdtable, args, self.middleware)
+
+        try:
+            name, func, args, kwargs = self._dispatch(args)
+        except Exception, e:
+            if exchandle(e, help_func):
+                return -1
+            raise
+
+        try:
+            if name == '_completion' or not self.middleware:
+                return call_cmd(name, func)(*args, **kwargs)
+            else:
+                return call_cmd(name, self.middleware(func), depth=2)(
+                    *args, **kwargs)
+        except Exception, e:
+            if exchandle(e, help_func):
+                return -1
+            raise
+
+
+_dispatcher = None
+
+def command(options=None, usage=None, name=None, shortlist=False, hide=False):
+    global _dispatcher
+    if not _dispatcher:
+        _dispatcher = Dispatcher()
+    return _dispatcher.command(options=options, usage=usage, name=name,
+                               shortlist=shortlist, hide=hide)
+command.__doc__ = Dispatcher.command.__doc__
+
+def dispatch(args=None, cmdtable=None, globaloptions=None, middleware=None):
+    global _dispatcher
+    if not _dispatcher:
+        _dispatcher = Dispatcher(cmdtable, globaloptions, middleware)
     else:
-        mwfunc = middleware(func)
-        depth = func == mwfunc and 1 or 2
-        worker = lambda: call_cmd(name, mwfunc, depth=depth)(*args, **kwargs)
+        if cmdtable:
+            _dispatcher.cmdtable = cmdtable
+        if globaloptions:
+            _dispatcher._globaloptions = globaloptions
+        if middleware:
+            _dispatcher.middleware = middleware
+    return _dispatcher.dispatch(args)
+dispatch.__doc__ = Dispatcher.dispatch.__doc__
 
-    try:
-        return catcher(worker, help_func)
-    except Abort:
-        return -1
 
 # --------
 # Help
@@ -203,7 +258,7 @@ def help_(cmdtable, globalopts):
                 if shortlist and not cmd.startswith('^'):
                     continue  # short help contains only marked commands
                 cmd = cmd.lstrip('^~')
-                doc = info[0].__doc__ or '(no help text available)'
+                doc = pretty_doc_string(info[0])
                 hlp[cmd] = doc.strip().splitlines()[0].rstrip()
 
             hlplist = sorted(hlp)
@@ -262,7 +317,7 @@ def help_cmd(func, usage, options):
         --pid-file   name of file to write process ID to
     '''
     write(usage + '\n')
-    doc = func.__doc__ or '(no help text available)'
+    doc = pretty_doc_string(func)
     write('\n' + doc.strip() + '\n\n')
     if options:
         write(''.join(help_options(options)))
@@ -323,7 +378,7 @@ def parse(args, options):
             raise OpsterError(
                 'Long name should be defined for every option')
         # change name to match Python styling
-        pyname = name.replace('-', '_')
+        pyname = name_to_python(name)
         argmap['-' + short] = argmap['--' + name] = pyname
         defmap[pyname] = default
 
@@ -380,18 +435,6 @@ def parse(args, options):
 # Subcommand system
 # --------
 
-def _dispatch(args, cmdtable, globalopts):
-    '''Dispatch arguments list by a command table
-    '''
-    cmd, func, args, options = cmdparse(args, cmdtable, globalopts)
-
-    if options.pop('help', False):
-        return 'help', cmdtable['help'][0], [cmd], {}
-    elif not cmd:
-        return 'help', cmdtable['help'][0], ['shortlist'], {}
-
-    return cmd, func, args, options
-
 def cmdparse(args, cmdtable, globalopts):
     '''Parse arguments list to find a command, options and arguments
     '''
@@ -413,12 +456,8 @@ def cmdparse(args, cmdtable, globalopts):
 
     possibleopts.extend(globalopts)
 
-    try:
-        options, args = parse(args, possibleopts)
-    except getopt.GetoptError, e:
-        raise ParseError(cmd, e)
-
-    return (cmd, cmd and info[0] or None, args, options)
+    options, args = parse(args, possibleopts)
+    return cmd, cmd and info[0] or None, args, options
 
 def aliases_(cmdtable_key):
     '''Get aliases from a command table key'''
@@ -481,49 +520,50 @@ def guess_options(func):
         try:
             sname, default, hlp = option[:3]
             completer = option[3] if len(option) > 3 else None
-            yield (sname, name.replace('_', '-'), default, hlp, completer)
+            yield (sname, name_from_python(name), default, hlp, completer)
         except TypeError:
             pass
 
 def guess_usage(func, options):
     '''Get usage definition for a function
     '''
-    usage = '%name '
+    usage = ['%name']
     if options:
-        usage += '[OPTIONS] '
-    args, varargs = inspect.getargspec(func)[:2]
-    argnum = len(args) - len(options)
-    if argnum > 0:
-        usage += args[0].upper()
-        if argnum > 1:
-            usage += 'S'
-    elif varargs:
-        usage += '[%s]' % varargs.upper()
-    return usage
+        usage.append('[OPTIONS]')
+    arginfo = inspect.getargspec(func)
 
-def catcher(target, help_func):
-    '''Catches all exceptions and prints human-readable information on them
+    argnum = len(arginfo.args or ()) - len(options)
+    nonoptional = len(arginfo.args or ()) - len(arginfo.defaults or ())
+
+    for i in xrange(argnum):
+        usage.append((i > nonoptional - 1 and '[%s]' or '%s') %
+                     arginfo.args[i].upper())
+
+    if arginfo.varargs:
+        usage.append('[%s ...]' % arginfo.varargs.upper())
+    return ' '.join(usage)
+
+def exchandle(e, help_func):
+    '''Handle internal exceptions and print human-readable information on them
+
+    Returns False is exception is not suitable
     '''
-    try:
-        return target()
-    except UnknownCommand, e:
+    if isinstance(e, UnknownCommand):
         err("unknown command: '%s'\n" % e)
-        raise Abort()
-    except AmbiguousCommand, e:
+    elif isinstance(e, AmbiguousCommand):
         err("command '%s' is ambiguous:\n    %s\n" %
             (e.args[0], ' '.join(e.args[1])))
-        raise Abort()
-    except ParseError, e:
+    elif isinstance(e, ParseError):
         err('%s: %s\n\n' % (e.args[0], e.args[1].strip()))
         help_func(e.args[0])
-        raise Abort()
-    except getopt.GetoptError, e:
+    elif isinstance(e, getopt.GetoptError):
         err('error: %s\n\n' % e)
         help_func()
-        raise Abort()
-    except OpsterError, e:
+    elif isinstance(e, OpsterError):
         err('%s\n' % e)
-        raise Abort()
+    else:
+        return False
+    return True
 
 def call_cmd(name, func, depth=1):
     '''Wrapper for command call, catching situation with insufficient arguments
@@ -543,14 +583,15 @@ def call_cmd_regular(func, opts):
     '''Wrapper for command for handling function calls from Python
     '''
     def inner(*args, **kwargs):
-        funcargs, _, varkw, defaults = inspect.getargspec(func)
-        if len(args) > len(funcargs):
+        arginfo = inspect.getargspec(func)
+        if len(args) > len(arginfo.args):
             raise TypeError('You have supplied more positional arguments'
                             ' than applicable')
 
-        funckwargs = dict((lname.replace('-', '_'), default)
-                          for _, lname, default, _ in opts)
-        if 'help' not in (defaults or ()) and not varkw:
+        # short name, long name, default, help, (maybe) completer
+        funckwargs = dict((o[1].replace('-', '_'), o[2])
+                          for o in opts)
+        if 'help' not in (arginfo.defaults or ()) and not arginfo.keywords:
             funckwargs.pop('help', None)
         funckwargs.update(kwargs)
         return func(*args, **funckwargs)
@@ -571,20 +612,25 @@ def sysname():
         return name[2:]
     return name
 
-try:
-    from functools import wraps
-except ImportError:
-    def wraps(wrapped, assigned=('__module__', '__name__', '__doc__'),
-              updated=('__dict__',)):
-        '''functools.wraps replacement for Python 2.4
-        '''
-        def inner(wrapper):
-            for attr in assigned:
-                setattr(wrapper, attr, getattr(wrapped, attr))
-            for attr in updated:
-                getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
-            return wrapper
-        return inner
+def pretty_doc_string(item):
+    "Doc string with adjusted indentation level of the 2nd line and beyond."
+    raw_doc = item.__doc__ or '(no help text available)'
+    lines = raw_doc.strip().splitlines()
+    if len(lines) <= 1:
+        return raw_doc
+    indent = len(lines[1]) - len(lines[1].lstrip())
+    return '\n'.join([lines[0]] + map(lambda l: l[indent:], lines[1:]))
+
+def name_from_python(name):
+    if name.endswith('_') and keyword.iskeyword(name[:-1]):
+        name = name[:-1]
+    return name.replace('_', '-')
+
+def name_to_python(name):
+    name = name.replace('-', '_')
+    if keyword.iskeyword(name):
+        return name + '_'
+    return name
 
 # --------
 # Autocomplete system
@@ -629,7 +675,9 @@ def autocomplete(cmdtable, args, middleware):
             options += [short, long]
 
             if cwords[idx] in (short, long) and completer:
-                args = middleware(completer)(current)
+                if middleware:
+                    completer = middleware(completer)
+                args = completer(current)
                 print ' '.join(args),
 
         print ' '.join((o for o in options if o.startswith(current)))
@@ -692,9 +740,6 @@ class UnknownCommand(OpsterError):
 class ParseError(OpsterError):
     'Raised on error in command line parsing'
 
-class Abort(OpsterError):
-    'Processing error, abort execution'
-
 if __name__ == '__main__':
     import doctest
     doctest.testmod()
